#ifndef GAMEENGINE_H
#define GAMEENGINE_H

#include <GL/glew.h>
#include <GLFW/glfw3.h>
#include <iostream>
#include <string>
#include <fstream>
#include <cstring>
#include "KeyManager.h"
#include "MouseManager.h"
#include "Shader.h"
#include "DirectRenderer.h"
#include "Clock.h"

using namespace std;

typedef void (*InitFunction) ();
typedef void (*TickFunction) ();
typedef void (*RenderTickFunction) (float);

class GameEngine
{
public:
    static void start(int width, int height, const char* name, InitFunction init, TickFunction tick, RenderTickFunction renderTick);
    static void stop();
    static GameEngine* get();
    
    virtual ~GameEngine();
    
    KeyManager& getKeyManager();
    MouseManager& getMouseManager();
    Shader& getShader();
    DirectRenderer& getDirectRenderer();
    
    void printError();
    
    double getTicksPerSecond() const;
    double getFramesPerSecond() const;
    
    int getScale() const;
    int getWidth() const;
    int getHeight() const;
    
    GLint getUniformLocation(const GLchar* name) const;
    void setMatrix(GLint location, const GLfloat* m);
    void setInt(GLint location, GLint i);
    void setFloat(GLint location, GLfloat f1, GLfloat f2, GLfloat f3, GLfloat f4);
private:
    GameEngine(GLuint program, GLuint vShader, GLuint fShader, GLFWwindow* window, int width, int height, TickFunction tick, RenderTickFunction renderTick);
    static GLuint setInstance(GLuint program, GameEngine* en);
    static GameEngine* instance;
    static const uint64_t NANOS_PER_TICK = 50000000;
    static const int MAX_TICKS_PER_FRAME = 5;
    void run();

    void onKeyEvent(GLFWwindow* w, int key, int scancode, int action, int mods);
    void onMouseClick(GLFWwindow* w, int button, int action, int mods);
    void onWindowResize(GLFWwindow* w, int width, int height);

    static GLchar* readFile(const char* name);
    static bool checkShaderErrors(const char* name, GLuint shader);
    static GLuint compileProgram(GLuint& vShader, GLuint& fShader, const GLchar* vertex, const GLchar* fragment);
    static GLuint createProgram(GLuint& vShader, GLuint& fShader);

    void updateScale();

    TickFunction tick;
    RenderTickFunction renderTick;
    
    Clock tps;
    Clock fps;
    
    GLFWwindow* window;
    GLuint program = 0;
    GLuint vShader = 0;
    GLuint fShader = 0;

    int scale = 1;
    int width;
    int height;
    
    KeyManager keyManager;
    MouseManager mouseManager;
    Shader shader;
    DirectRenderer directRenderer;
};

#endif

